// HTTPServer runs asynchronously, accepting connections
// and passing them to the provided HTTPConnection handler

#include "HTTPServer.h"

#include "HTTPConnection.h"


// Construct
HTTPServer::HTTPServer(
	ProbabilityMap *map,
	RobotHandler *robotHandler,
	ImageProvider *imgProvider,
	HTTPConnection *connectionHandler
) :
	ArASyncTask(),
	
	m_map(map),
	m_robotHandler(robotHandler),
	m_imgProvider(imgProvider),
	m_connectionHandler(connectionHandler)
{
}


// Main server loop
void* HTTPServer::runThread(void*) {
	// Listen
	//m_sock.setReuseAddress();
	if(m_sock.open(80, ArSocket::TCP)) {
		m_sock.setDoClose(true);
		
		// Loop
		while(getRunningWithLock()) {
			ArSocket	*sock = new ArSocket;
			
			// Wait for incoming connection
			if(m_sock.accept(sock)) {
				bool handled = false;
				
				// Tell the connection handler to handle the connection
				handled = m_connectionHandler->handleConnection(sock);
				
				if(!handled) {
					sock->close();
					delete sock;
				}
			}
			else {
				delete sock;
				
				break;
			}
		}
		
		m_sock.close();
	}
	
	return NULL;
}

// Close the server socket
void HTTPServer::myStop() {
	m_sock.close();
}
