// JPEGLoader.cpp

// Decodes JPEG files
// Should be replaced with JPEG-library, like libjpeg or something,
// if it can be easily integrated and cross-platform
// Developed 50/50 on VC++2008 Vista and Linux on the Robot so a single source-file
// was easier

// (The actual code is rather un-explained since JPEG parsing is too complex to
// describe in code-comments, and rather irrelevant)

////////!!!!!!!////
// It is also not perfect, some images might fail to load
// JPEG images can have very different layout and only those that the
// particular used camera on the particular used setting provides are tested thoroughly
///////////////

#include "JPEGLoader.h"
#include <cmath>

// Stuff used in decoding JPEG
struct DataInfo {
	const unsigned char		*data;
	int						len;
	
	unsigned char			*pixels;
	int						width;
	int						iwidth;
	int						height;
	
	int						*dcodes[2];
	int						*acodes[2];
	int						*dtable[2];
	int						*atable[2];
	
	int						*quant[3];
	
	int						*hsamp;
	int						*vsamp;
	int						*qnum;
	
	int						resinterval;
	
	DataInfo() :
		data(0),
		len(0),
		
		pixels(0),
		width(0),
		iwidth(0),
		height(0),
		
		hsamp(0),
		vsamp(0),
		qnum(0),
		
		resinterval(0)
	{
		for(int i=0;i<2;i++) {
			dcodes[i] = acodes[i] = dtable[i] = atable[i] = 0;
		}
		
		for(int i=0;i<3;i++) {
			quant[i] = 0;
		}
	}
};


bool readJpeg(DataInfo *img);
bool readScan(DataInfo *img, int ibyte, int totlen);


// Load JPEG from data
// pixelsa, widtha, iwidtha, heighta are output image pixels and dimensions
// The image will always be stored with a width of a multiple of 4 pixels, the value
// stored in *iwidtha, while *widtha will be the actual used image width
bool loadJPEG(const void *vdata, int len, unsigned char **pixelsa, int *widtha, int *iwidtha, int *heighta) {
	DataInfo img;
	
	img.data = static_cast<const unsigned char*>(vdata);
	img.len = len;
	img.pixels = 0;
	
	if(readJpeg(&img)) {
		*pixelsa = img.pixels;
		*widtha = img.width;
		*iwidtha = img.iwidth;
		*heighta = img.height;
		
		return true;
	}
	else {
		delete[] img.pixels;
		
		return false;
	}
}// loadJPEG


// Build codes (for JPEG decoding)
void jbuildCodes(int bits[16], int *bytes, int *vals) {
	int sum = 0;
	int tat = 1;
	vals[0] = -1;
	for(int depth=0;depth<16;depth++) {
		for(int i=0;i<bits[depth];i++)
			vals[tat++] = bytes[sum++];
		
		tat <<= 1;
		tat++;
	}
}// jbuildCodes


// Make codewords 32-bit
void makeCodes32(int *vals, int *codes, int *codeLens, int &nCodes, int code, int add, int len) {
	if(vals[0] >= 0) {
		codes[nCodes] = code;
		codeLens[nCodes] = len;
		nCodes++;
	}
	else if(add < 100000) {
		makeCodes32(vals+add, codes, codeLens, nCodes, (code<<1), (add<<1), len+1);
		makeCodes32(vals+add+1, codes, codeLens, nCodes, (code<<1)|0x1, (add<<1)+1, len+1);
	}
}// makeCodes


// Build lookup table for codes
void buildTable(int ncodes, int *codeLens, int *codes, int *bytes, int *table) {
	for(int i=0;i<ncodes;i++) {
		int codeLen = codeLens[i];
		int code = codes[i];
		int byte = bytes[i];
		
		int num = (code << (16 - codeLen));
		
		for(int j=0;j<(1 << (16 - codeLen));j++) {
			int tableIndex = num + j;
			
			table[tableIndex] = (codeLen << 8) | byte;
		}
	}
}// buildTable


// Read JPEG
bool readJpeg(DataInfo *img) {
	const unsigned char *data = img->data;
	int len = img->len;
	
	bool success = false;
	
	bool hasImg = false;
	bool hasApp0 = false;
	int quantCount = 0;
	bool hasRes = false;
	int acTableCount = 0;
	int dcTableCount = 0;
	bool hasInfo = false;
	
	bool bail = false;
	
	int dc[1024];
	int dcount[32];
	int ac[1024];
	int acount[32];
	
	int hsamp[3];
	int vsamp[3];
	int qnum[3];
	
	int *dcodes0 = new int[131072];
	int *dcodes1 = new int[131072];
	int *acodes0 = new int[131072];
	int *acodes1 = new int[131072];
	
	for(int i=0;i<131072;i++) {
		dcodes0[i] = -1;
		dcodes1[i] = -1;
		acodes0[i] = -1;
		acodes1[i] = -1;
	}
	
	int quant[256];
	int resinterval = 0;
	
	int mlen;
	
	// Parse JPEG sections
	for(int i=1;i<len&&!bail;i++) {
		if(data[i-1] == 0xff) {
			// Image start
			if(data[i] == 0xd8) {
				hasImg = true;
			}
			
			// JFIF
			else if(data[i] == 0xe0 && i < len-2) {
				mlen = (int)*(data+i + 1) * 256 + (int)*(data+i + 2);
				
				if(hasImg) {
					hasApp0 = true;
				}
				
				i += mlen;
			}
			
			// Quantization table(s)
			else if(data[i] == 0xdb && i < len-2) {
				mlen = (int)*(data+i + 1) * 256 + (int)*(data+i + 2);
				
				if(hasApp0 && i+mlen < len) {
					int tot = 3;
					while(tot <= mlen-64) {
						int b = (int)*(data+i + tot);
						tot += 1;
						//int prec = (b >> 4) & 0xf;
						int index = b & 0xf;
						
						if(index == 0 || index == 1 || index == 2) {
							for(int j=0;j<64;j++) {
								quant[index*64+j] = (int)*(data+i+tot + j);
							}
							tot += 64;
							
							quantCount++;
						}
						else {
							bail = true;
							break;
						}
					}
				}
				
				i += mlen;
			}
			
			// Restart interval
			else if(data[i] == 0xdd && i < len-2) {
				mlen = (int)*(data+i + 1) * 256 + (int)*(data+i + 2);
				
				if(hasApp0) {
					if(mlen == 4 && i+mlen < len) {
						resinterval = (int)*(data+i + 3) * 256 + (int)*(data+i + 4);
						
						hasRes = true;
					}
				}
				
				i += mlen;
			}
			
			// Huffman
			else if(data[i] == 0xc4 && i < len-2) {
				mlen = (int)*(data+i + 1) * 256 + (int)*(data+i + 2);
				
				if(hasApp0 && i+mlen < len) {
					int tot = 3;
					while(tot < mlen-17) {
						int index = (int)*(data+i + tot);
						tot += 1;
						int bits = 0;
						
						if(index >= 0x10) {
							index -= 0x10;
							
							if(acTableCount >= 2 || (index != 0 && index != 1)) { bail = true; break; }
							
							for(int j=0;j<16;j++) {
								acount[index*16+j] = (int)*(data+i+tot + j);
								bits += (int)*(data+i+tot + j);
							}
							tot += 16;
							
							if(i+tot+bits >= len) { bail = true; break; }
							
							for(int j=0;j<bits;j++)
								ac[index*512+j] = (int)*(data+i+tot + j);
							tot += bits;
							
							acTableCount++;
						}
						else {
							if(dcTableCount >= 2 || (index != 0 && index != 1)) { bail = true; break; }
							
							for(int j=0;j<16;j++) {
								dcount[index*16+j] = (int)*(data+i+tot + j);
								bits += (int)*(data+i+tot + j);
							}
							tot += 16;
							
							if(i+tot+bits >= len) { bail = true; break; }
							
							for(int j=0;j<bits;j++)
								dc[index*512+j] = (int)*(data+i+tot + j);
							tot += bits;
							
							dcTableCount++;
						}
					}
				}
				
				i += mlen;
			}
			
			// Image info
			else if(data[i] == 0xc0 && i < len-2) {
				mlen = (int)*(data+i + 1) * 256 + (int)*(data+i + 2);
				
				if(hasApp0 && i+mlen < len && mlen >= 17) {
					int p = (int)*(data+i + 3);
					int y = (int)*(data+i + 4) * 256 + (int)*(data+i + 5);
					int x = (int)*(data+i + 6) * 256 + (int)*(data+i + 7);
					int nf = (int)*(data+i + 8);
					
					if(p == 8 && x > 0 && y > 0 && nf == 3) {
						img->width = x;
						img->height = y;
						
						img->iwidth = img->width;
						if((img->iwidth % 4) != 0) {
							img->iwidth += (4 - (img->iwidth % 4));
						}
						
						delete[] img->pixels;
						img->pixels = new unsigned char[img->height*img->iwidth*3];
						
						for(int j=0;j<nf;j++) {
							//int id = (int)*(data+i+9 + j*3 + 0);
							int hv = (int)*(data+i+9 + j*3 + 1);
							int h = (hv >> 4) & 0xf;
							int v = hv & 0xf;
							int qn = (int)*(data+i+9 + j*3 + 2);
							
							hsamp[j] = h;
							vsamp[j] = v;
							qnum[j] = qn;
						}
						
						hasInfo = true;
					}
				}
				
				i += mlen;
			}
			
			// Scan start
			else if(data[i] == 0xda && i < len-2) {
				mlen = (int)*(data+i + 1) * 256 + (int)*(data+i + 2);
				
				if(hasInfo && dcTableCount == 2 && acTableCount == 2 && quantCount >= 2) {
					jbuildCodes(dcount, dc, dcodes0);
					jbuildCodes(dcount+16, dc+512, dcodes1);
					jbuildCodes(acount, ac, acodes0);
					jbuildCodes(acount+16, ac+512, acodes1);
					
					int *dtable0 = new int[65536];
					int *dtable1 = new int[65536];
					int *atable0 = new int[65536];
					int *atable1 = new int[65536];
					
					int codeLens[512];
					int codes[512];
					int nCodes;
					
					int tot;
					
					tot = 0;
					for(int j=0;j<16;j++)
						tot += dcount[j];
					nCodes = 0;
					makeCodes32(dcodes0, codes, codeLens, nCodes, 0, 1, 0);
					buildTable(tot, codeLens, codes, dc, dtable0);
					
					tot = 0;
					for(int j=0;j<16;j++)
						tot += dcount[16+j];
					nCodes = 0;
					makeCodes32(dcodes1, codes, codeLens, nCodes, 0, 1, 0);
					buildTable(tot, codeLens, codes, dc+512, dtable1);
					
					tot = 0;
					for(int j=0;j<16;j++)
						tot += acount[j];
					nCodes = 0;
					makeCodes32(acodes0, codes, codeLens, nCodes, 0, 1, 0);
					buildTable(tot, codeLens, codes, ac, atable0);
					
					tot = 0;
					for(int j=0;j<16;j++)
						tot += acount[16+j];
					nCodes = 0;
					makeCodes32(acodes1, codes, codeLens, nCodes, 0, 1, 0);
					buildTable(tot, codeLens, codes, ac+512, atable1);
					
					
					img->dcodes[0] = dcodes0;
					img->dcodes[1] = dcodes1;
					img->acodes[0] = acodes0;
					img->acodes[1] = acodes1;
					
					img->quant[0] = quant;
					img->quant[1] = (quant + 64);
					img->quant[2] = (quant + 128);
					
					img->resinterval = resinterval;
					img->hsamp = hsamp;
					img->vsamp = vsamp;
					img->qnum = qnum;
					
					img->dtable[0] = dtable0;
					img->dtable[1] = dtable1;
					img->atable[0] = atable0;
					img->atable[1] = atable1;
					
					// Read the actual image
					success = readScan(img, i+mlen+1, len-i-mlen-1);
					
					delete[] dtable0;
					delete[] dtable1;
					delete[] atable0;
					delete[] atable1;
					
					break;
				}
				
				i += mlen;
			}
			
			// Other
			else if(data[i] != 0xff && i < len-2) {
				mlen = (int)*(data+i + 1) * 256 + (int)*(data+i + 2);
				
				i += mlen;
			}
			
			else {
				//bail = true;
				//break;
			}
		}
	}
	
	delete[] dcodes0;
	delete[] dcodes1;
	delete[] acodes0;
	delete[] acodes1;
	
	return success;
}// readJpeg


// Restart
inline void doRestart2(unsigned int &dword, int &numThis, unsigned int* &bitArray, int &counter, int len16, bool &bail) {
	if((numThis%8) != 0) {
		int numNew = (numThis%8);
		
		dword <<= numNew;
		numThis -= numNew;
		
		if(numThis < 16) {
			counter++;
			if(counter >= len16) {
				bail = true;
				return;
			}
			
			dword |= ((*bitArray++) << (16 - numThis));
			numThis += 16;
		}
	}
}// doRestart2


// Decode DC
inline int decodedc322(unsigned int &dword, int &numThis, unsigned int* &bitArray, int *codetable, int &counter, int len16, bool &bail) {
	int bits = codetable[dword>>16];
	int codeLen = bits >> 8;
	int valLen = bits & 0xff;
	
	dword <<= codeLen;
	numThis -= codeLen;
	
	if(valLen == 0) {
		if(numThis < 16) {
			counter++;
			if(counter >= len16) {
				bail = true;
				return 0;
			}
			
			dword |= ((*bitArray++) << (16 - numThis));
			numThis += 16;
		}
		
		return 0;
	}
	else {
		if(numThis < valLen) {
			counter++;
			if(counter >= len16) {
				bail = true;
				return 0;
			}
			
			dword |= ((*bitArray++) << (16 - numThis));
			numThis += 16;
		}
		
		int shiftLen = (32 - valLen);
		
		int dval = (dword >> shiftLen);
		
		int cp = (((~dval) >> (valLen-1)) & 1);
		dval -= (cp << valLen) - cp;
		
		dword <<= valLen;
		numThis -= valLen;
		
		if(numThis < 16) {
			counter++;
			if(counter >= len16) {
				bail = true;
				return 0;
			}
			
			dword |= ((*bitArray++) << (16 - numThis));
			numThis += 16;
		}
		
		return dval;
	}
}// decodedc322


// Decode AC
inline int decodeac322(unsigned int &dword, int &numThis, unsigned int* &bitArray, int *codetable, int &zrlr, int &counter, int len16, bool &bail) {
	int bits = codetable[dword>>16];
	int codeLen = bits >> 8;
	int valLen = bits & 0xff;
	
	dword <<= codeLen;
	numThis -= codeLen;
	
	zrlr = (valLen >> 4);
	valLen &= 0x0f;
	
	if(valLen == 0) {
		if(numThis < 16) {
			counter++;
			if(counter >= len16) {
				bail = true;
				return 0;
			}
			
			dword |= ((*bitArray++) << (16 - numThis));
			numThis += 16;
		}
		
		return 0;
	}
	else {
		if(numThis < valLen) {
			counter++;
			if(counter >= len16) {
				bail = true;
				return 0;
			}
			
			dword |= ((*bitArray++) << (16 - numThis));
			numThis += 16;
		}
		
		int shiftLen = (32 - valLen);
		int cp = ((~dword) >> 31);
		
		int dval = (dword >> shiftLen);
		
		dval -= (cp << valLen) - cp;
		
		dword <<= valLen;
		numThis -= valLen;
		
		if(numThis < 16) {
			counter++;
			if(counter >= len16) {
				bail = true;
				return 0;
			}
			
			dword |= ((*bitArray++) << (16 - numThis));
			numThis += 16;
		}
		
		return dval;
	}
}// decodeac322


// Read scan
bool readScan(DataInfo *img, int ibyte, int totlen) {
	if(totlen < 100) return false;
	
	const unsigned char *data = img->data;
	int len = img->len;
	
	//int blen = (len-ibyte+1)*8;
	
	if(ibyte > len-100) return false;
	
	// Extract 16bit words
	unsigned int *bits16 = new unsigned int[len-ibyte];
	unsigned int *bytes = new unsigned int[len-ibyte];
	int bylen = 0;
	//int bilen = 0;
	for(int j=ibyte;j<len;j++) {
		unsigned int b = (unsigned int)data[j];
		
		if(b == 0xff) {
			if(j < len-1) {
				if(data[j+1] != 0x00 && data[j+1] != 0xff) {
					// Restart marker
					if(data[j+1] >= 0xd0 && data[j+1] <= 0xd7) {
						j++;
						continue;
					}
					else {
						// End
						break;
					}
				}
				else// 0xff in data
					j++;
			}
			else
				continue;
		}
		
		bytes[bylen++] = b;
	}
	int len16 = 0;
	for(int j=0;j<bylen;j+=2) {
		unsigned int bi = bytes[j] << 8;
		
		if(j+1 < bylen) bi |= bytes[j+1];
		
		bits16[j/2] = bi;
		
		if(j+2 >= bylen) {
			len16 = (j/2) + 1 + 2;
			bits16[len16-1] = 0;
			bits16[len16-2] = 0;
		}
	}
	
	// Build IDCT table
	double costable[64];
	for(int s=0;s<8;s++) {
		for(int t=0;t<8;t++) {
			costable[(s<<3)+t] = std::cos((2.0*static_cast<double>(s)+1.0)*static_cast<double>(t)*3.14159265358979/16.0);
		}
	}
	float fcostable[64];
	for(int x=0;x<8;x++) {
		for(int u=0;u<8;u++) {
			double Cu = 1.0;
			if(u == 0)
				Cu = 1.0 / std::sqrt(2.0);
			
			fcostable[(x<<3)+u] = static_cast<float>(0.5 * Cu * costable[(x<<3)+u]);
		}
	}
	
	// Keep track
	int reldc[3];
	
	for(int j=0;j<3;j++)
		reldc[j] = 0;
	
	int xstart = 0;
	int ystart = 0;
	
	int imcu = 0;
	
	float RVals[768];
	float FVals[768];
	
	bool	bail = false;
	
	if(len16 < 10) {
		bail = true;
		//goto bailLoop;
	}
	
	////////
	unsigned int *cpbits = bits16;
	unsigned int dword = 0;//
	int numThis = 0;
	if(!bail) {
		dword = (cpbits[0] << 16) | cpbits[1];
		numThis = 32;
		cpbits += 2;
	}
	//////////
	
	for(int i=0;i<3;i++) {
		if(img->qnum[i] < 0 || img->qnum[i] > 2) {
			bail = true;
		}
	}
	
	// Read image
	for(int i=2;i<len16&&!bail;) {
		// Restarts
		if(img->resinterval > 0) {
			if(imcu >= img->resinterval) {
				imcu = 0;
				for(int j=0;j<3;j++)
					reldc[j] = 0;
				
				doRestart2(dword, numThis, cpbits, i, len16, bail);
				if(bail) goto bailLoop;
			}
			imcu++;
		}
		
		// Read channels
		for(int j=0;j<3;j++) {
			int dtab = 0;
			if(j != 0) dtab = 1;
			int atab = 0;
			if(j != 0) atab = 1;
			
			int val;
			
			int nSamp = img->hsamp[j] * img->vsamp[j];
			
			if(nSamp > 4) {
				bail = true;
				goto bailLoop;
			}
			
			for(int ip=0;ip<nSamp;ip++) {
				float *RVal = (RVals + j*256 + ip*64);
				float *FVal = (FVals + j*256 + ip*64);
				
				// Decode and quantize
				int *quant = img->quant[img->qnum[j]];
				
				int zrl;
				
				val = decodedc322(dword, numThis, cpbits, img->dtable[dtab], i, len16, bail);
				if(bail) goto bailColor;
				
				reldc[j] += val;
				RVal[0] = (float)(reldc[j] * quant[0]);
				
				static const int zind[64] = {
					0, 1, 8, 16, 9, 2, 3, 10, 17, 24, 32, 25, 18, 11, 4, 5,
					12, 19, 26, 33, 40, 48, 41, 34, 27, 20, 13, 6, 7, 14, 21, 28,
					35, 42, 49, 56, 57, 50, 43, 36, 29, 22, 15, 23, 30, 37, 44, 51,
					58, 59, 52, 45, 38, 31, 39, 46, 53, 60, 61, 54, 47, 55, 62, 63
				};
				
				int k = 1;
				for(;k<64;k++) {
					val = decodeac322(dword, numThis, cpbits, img->atable[atab], zrl, i, len16, bail);
					if(bail) goto bailColor;
					
					if(zrl == 0 && val == 0) {
						break;
					}
					
					for(int l=0;l<zrl;l++)
						RVal[zind[k++]] = 0.0f;
					RVal[zind[k]] = (float)(val * quant[k]);
				}
				for(;k<64;k++)
					RVal[zind[k]] = 0.0f;
				
				// Transform
				float *rv = RVal;
				for(int y=0;y<8;y++) {
					for(int x=0;x<8;x++) {
						float *fc = (fcostable + x*8);
						float cc = 0.0f;
						
						for(int u=0;u<8;u++)
							cc += rv[u] * fc[u];
						
						FVal[y+x*8] = cc;
					}
					
					rv += 8;
				}
				
				float *fv = FVal;
				for(int x=0;x<8;x++) {
					for(int y=0;y<8;y++) {
						float *fc = (fcostable + y*8);
						float cc = 128.5f;
						
						for(int v=0;v<8;v++)
							cc += fv[v] * fc[v];
						
						RVal[y*8+x] = cc;
					}
					
					fv += 8;
				}
			}
		}
		
		// Store RGB
		// Handles a few different cases depending on sub-sampling of chroma channels
		bailColor:
		
		float *FY = RVals;
		float *FCb = RVals + 256;
		float *FCr = RVals + 512;
		
		if(img->vsamp[0] == 1 && img->hsamp[0] == 1) {
			int h = 8;
			if(h > img->height-ystart) h = img->height-ystart;
			int w = 8;
			if(w > img->width-xstart) w = img->width-xstart;
			
			unsigned char *spix = (img->pixels + ystart*img->iwidth*3 + xstart*3);
			for(int y=0;y<h;y++) {
				unsigned char *pix = (spix + y*img->iwidth*3);
				
				for(int x=0;x<w;x++) {
					float cy = (float)(int)FY[x];
					float ccb = (float)(int)FCb[x];
					float ccr = (float)(int)FCr[x];
					
					if(cy<0)cy=0;if(cy>255)cy=255;
					if(ccb<0)ccb=0;if(ccb>255)ccb=255;
					if(ccr<0)ccr=0;if(ccr>255)ccr=255;
					
					int cr = (int)(cy + 1.402f * (ccr - 128.0f) + 0.5f);
					int cg = (int)(cy - 0.34414f * (ccb - 128.0f) - 0.71414f * (ccr - 128.0f) + 0.5f);
					int cb = (int)(cy + 1.772f * (ccb - 128.0f) + 0.5f);
					
					if(cr<0)cr=0;if(cr>255)cr=255;
					if(cg<0)cg=0;if(cg>255)cg=255;
					if(cb<0)cb=0;if(cb>255)cb=255;
					
					pix[x*3] = (unsigned char)cb;
					pix[x*3+1] = (unsigned char)cg;
					pix[x*3+2] = (unsigned char)cr;
				}
				FY += 8;
				FCb += 8;
				FCr += 8;
			}
		}
		else if(img->vsamp[0] == 1 && img->hsamp[0] == 2) {
			int h = 8;
			if(h > img->height-ystart) h = img->height-ystart;
			int w = 16;
			if(w > img->width-xstart) w = img->width-xstart;
			
			unsigned char *spix = (img->pixels + ystart*img->iwidth*3 + xstart*3);
			for(int y=0;y<h;y++) {
				unsigned char *pix = (spix + y*img->iwidth*3);
				
				for(int x=0;x<w;x++) {
					float cy = (float)(int)FY[x];
					float ccb = (float)(int)FCb[x/2];
					float ccr = (float)(int)FCr[x/2];
					
					if(cy<0)cy=0;if(cy>255)cy=255;
					if(ccb<0)ccb=0;if(ccb>255)ccb=255;
					if(ccr<0)ccr=0;if(ccr>255)ccr=255;
					
					float rp = 1.402f * (ccr - 128.0f) + 0.5f;
					float gp = -0.34414f * (ccb - 128.0f) - 0.71414f * (ccr - 128.0f) + 0.5f;
					float bp = 1.772f * (ccb - 128.0f) + 0.5f;
					
					int cr = (int)(cy + rp);
					int cg = (int)(cy + gp);
					int cb = (int)(cy + bp);
					
					if(cr<0)cr=0;if(cr>255)cr=255;
					if(cg<0)cg=0;if(cg>255)cg=255;
					if(cb<0)cb=0;if(cb>255)cb=255;
					
					pix[x*3] = (unsigned char)cb;
					pix[x*3+1] = (unsigned char)cg;
					pix[x*3+2] = (unsigned char)cr;
					
					x++;
					
					cy = (float)(int)FY[x];
					
					if(cy<0)cy=0;if(cy>255)cy=255;
					
					cr = (int)(cy + rp);
					cg = (int)(cy + gp);
					cb = (int)(cy + bp);
					
					if(cr<0)cr=0;if(cr>255)cr=255;
					if(cg<0)cg=0;if(cg>255)cg=255;
					if(cb<0)cb=0;if(cb>255)cb=255;
					
					pix[x*3] = (unsigned char)cb;
					pix[x*3+1] = (unsigned char)cg;
					pix[x*3+2] = (unsigned char)cr;
					
					if(x == 7) {
						FY += 56;
					}
				}
				FY -= 56;
				FY += 8;
				
				FCb += 8;
				FCr += 8;
			}
		}
		else if(img->vsamp[0] == 2 && img->hsamp[0] == 1) {
			int h = 16;
			if(h > img->height-ystart) h = img->height-ystart;
			int w = 8;
			if(w > img->width-xstart) w = img->width-xstart;
			
			unsigned char *spix = (img->pixels + ystart*img->iwidth*3 + xstart*3);
			for(int y=0;y<h;y++) {
				unsigned char *pix = (spix + y*img->iwidth*3);
				
				for(int x=0;x<w;x++) {
					float cy = (float)(int)FY[x];
					float ccb = (float)(int)FCb[x];
					float ccr = (float)(int)FCr[x];
					
					if(cy<0)cy=0;if(cy>255)cy=255;
					if(ccb<0)ccb=0;if(ccb>255)ccb=255;
					if(ccr<0)ccr=0;if(ccr>255)ccr=255;
					
					float rp = 1.402f * (ccr - 128.0f) + 0.5f;
					float gp = -0.34414f * (ccb - 128.0f) - 0.71414f * (ccr - 128.0f) + 0.5f;
					float bp = 1.772f * (ccb - 128.0f) + 0.5f;
					
					int cr = (int)(cy + rp);
					int cg = (int)(cy + gp);
					int cb = (int)(cy + bp);
					
					if(cr<0)cr=0;if(cr>255)cr=255;
					if(cg<0)cg=0;if(cg>255)cg=255;
					if(cb<0)cb=0;if(cb>255)cb=255;
					
					pix[x*3] = (unsigned char)cb;
					pix[x*3+1] = (unsigned char)cg;
					pix[x*3+2] = (unsigned char)cr;
				}
				FY += 8;
				
				if(y&0x1) {
					FCb += 8;
					FCr += 8;
				}
			}
		}
		else if(img->vsamp[0] == 2 && img->hsamp[0] == 2) {
			int h = 16;
			if(h > img->height-ystart) h = img->height-ystart;
			int w = 16;
			if(w > img->width-xstart) w = img->width-xstart;
			
			unsigned char *spix = (img->pixels + ystart*img->iwidth*3 + xstart*3);
			for(int y=0;y<h;y++) {
				unsigned char *pix = (spix + y*img->iwidth*3);
				
				for(int x=0;x<w;x++) {
					float cy = (float)(int)FY[x];
					float ccb = (float)(int)FCb[x/2];
					float ccr = (float)(int)FCr[x/2];
					
					if(cy<0)cy=0;if(cy>255)cy=255;
					if(ccb<0)ccb=0;if(ccb>255)ccb=255;
					if(ccr<0)ccr=0;if(ccr>255)ccr=255;
					
					float rp = 1.402f * (ccr - 128.0f) + 0.5f;
					float gp = -0.34414f * (ccb - 128.0f) - 0.71414f * (ccr - 128.0f) + 0.5f;
					float bp = 1.772f * (ccb - 128.0f) + 0.5f;
					
					int cr = (int)(cy + rp);
					int cg = (int)(cy + gp);
					int cb = (int)(cy + bp);
					
					if(cr<0)cr=0;if(cr>255)cr=255;
					if(cg<0)cg=0;if(cg>255)cg=255;
					if(cb<0)cb=0;if(cb>255)cb=255;
					
					pix[x*3] = (unsigned char)cb;
					pix[x*3+1] = (unsigned char)cg;
					pix[x*3+2] = (unsigned char)cr;
					
					x++;
					if(x >= w) break;
					
					cy = (float)(int)FY[x];
					
					if(cy<0)cy=0;if(cy>255)cy=255;
					
					cr = (int)(cy + rp);
					cg = (int)(cy + gp);
					cb = (int)(cy + bp);
					
					if(cr<0)cr=0;if(cr>255)cr=255;
					if(cg<0)cg=0;if(cg>255)cg=255;
					if(cb<0)cb=0;if(cb>255)cb=255;
					
					pix[x*3] = (unsigned char)cb;
					pix[x*3+1] = (unsigned char)cg;
					pix[x*3+2] = (unsigned char)cr;
					
					if(x == 7) {
						FY += 56;
					}
				}
				if(w > 7)
					FY -= 56;
				FY += 8;
				
				if(y&0x1) {
					FCb += 8;
					FCr += 8;
				}
				if(y == 7) {
					FY += 64;
				}
			}
		}
		else {
			for(int iy=0;iy<img->vsamp[0];iy++) {
				for(int ix=0;ix<img->hsamp[0];ix++) {
					for(int y=0;y<8;y++) {
						if(ystart+(y+iy*8) > img->height-1) break;
						
						for(int x=0;x<8;x++) {
							if(xstart+(x+ix*8) > img->width-1) break;
							
							float cy = (float)((int)FY[(iy*img->hsamp[0]+ix)*64+y*8+x]);
							float ccb = (float)((int)FCb[((iy*8+y)/img->vsamp[0])*8+((ix*8+x)/img->hsamp[0])]);
							float ccr = (float)((int)FCr[((iy*8+y)/img->vsamp[0])*8+((ix*8+x)/img->hsamp[0])]);
							
							if(cy<0)cy=0;if(cy>255)cy=255;
							if(ccb<0)ccb=0;if(ccb>255)ccb=255;
							if(ccr<0)ccr=0;if(ccr>255)ccr=255;
							
							int cr = (int)(cy + 1.402 * (ccr - 128.0) + 0.5);
							int cg = (int)(cy - 0.34414 * (ccb - 128.0) - 0.71414 * (ccr - 128.0) + 0.5);
							int cb = (int)(cy + 1.772 * (ccb - 128.0) + 0.5);
							
							if(cr<0)cr=0;if(cr>255)cr=255;
							if(cg<0)cg=0;if(cg>255)cg=255;
							if(cb<0)cb=0;if(cb>255)cb=255;
							
							img->pixels[ystart*img->iwidth*3+xstart*3+(y+iy*8)*img->iwidth*3+(x+ix*8)*3+0] = (unsigned char)cb;
							img->pixels[ystart*img->iwidth*3+xstart*3+(y+iy*8)*img->iwidth*3+(x+ix*8)*3+1] = (unsigned char)cg;
							img->pixels[ystart*img->iwidth*3+xstart*3+(y+iy*8)*img->iwidth*3+(x+ix*8)*3+2] = (unsigned char)cr;
						}
					}
				}
			}
		}
		
		xstart += 8*img->hsamp[0];
		if(xstart > img->width-1) {
			xstart = 0;
			ystart += 8*img->vsamp[0];
		}
		if(ystart > img->height-1) {
			break;
		}
	}
	
	bailLoop:
	delete[] bytes;
	delete[] bits16;
	
	return true;
}// readScan
