// LocalizationMap.h

#ifndef My__LocalizationMap
#define My__LocalizationMap


// Dependencies
#include "vec2.h"
#include "Ray.h"
#include <vector>

class CamImage;


// Map line segment
struct Line {
	Line			*prev;
	Line			*next;
	line2d			line;
	
	inline Line() {}
	inline Line(const vec2d &p0, const vec2d &p1) : line(p0, p1) {}
	inline Line(const line2d &l) : line(l) {}
};


// LocalizationMap class
class LocalizationMap {
	public:
		LocalizationMap();
		~LocalizationMap();
		
		bool input(const vec2d &pos, const Ray *rays, int nRays);
		
		void drawGrayscale(unsigned char *data, int w, int h, double scale);
		
		double getTheta();
	
	private:
		void findLines();
		
		void drawLine(unsigned char *data, int w, int h, const line2d &seg);
		
		void projectImage(const vec2d &pos, const CamImage &image);
	
	
	private:
		std::vector<Line*>		m_segments;
		
		unsigned char			*m_pixels;
		int						m_width;
		int						m_height;
		
		vec2d					*m_points;
		int						m_nPoints;
		int						m_nMaxPoints;
		
		unsigned int			*m_lvals;
		int						m_lwidth;
		int						m_lheight;
		double					m_lrscale;
		double					*m_ctable;
		double					*m_stable;
		
		double					m_lastt;
};


#endif
