// MyActionLimit.cpp

// This action limits the movement of the robot depending on
// obstacles as recorded in the map, or dynamic objects detected
// from the camera image

#include "MyActionLimit.h"
#include "ProbabilityMap.h"
#include "RobotHandler.h"
#include <cmath>



// Construct
MyActionLimit::MyActionLimit(ProbabilityMap *map, RobotHandler *robotHandler, double mapScale) :
	ArAction("limit speed"),
	
	m_map(map),
	m_robotHandler(robotHandler),
	m_mapScale(mapScale),
	
	m_speed(200.0)
{
}// MyActionLimit::MyActionLimit


// Destruct
MyActionLimit::~MyActionLimit() {
}// MyActionLimit::~MyActionLimit


// Fire action
ArActionDesired* MyActionLimit::fire(ArActionDesired currentDesired) {
	m_desired.reset();
	
	// Stop if the path in front of the robot is blocked in the map
	ArPose pose = myRobot->getPose();
	vec2d pos = vec2d(pose.getX(), pose.getY()) * m_mapScale + m_map->mapCenter();
	
	vec2d dir = vec2d(std::cos(pose.getThRad()), std::sin(pose.getThRad()));
	
	if(m_map->isBlocked(pos, dir, 15, 3)) {
		m_desired.setVel(0.0);
	}
	
	// Stop if there are dynamic objects in front of the robot
	std::vector<Actor> &actors = m_robotHandler->getActors();
	
	pos = vec2d(pose.getX(), pose.getY());
	dir;
	
	for(size_t i=0;i<actors.size();++i) {
		Actor &a = actors[i];
		
		vec2d adir = (a.pos - pos);
		double dist = adir.length();
		double d = dot(adir/dist, dir);
		
		// 80 cm, ~17 degrees safety
		// This should be done differently, creating an unsafe area where
		// the robot stops or turns
		// It was previously handled by putting dynamic objects in the map temporarily,
		// and pathfinding around them, but to decouple the map completely from
		// dynamic object avoidance this is used instead
		if(dist < 800.0 && d < 0.3) {
			m_desired.setVel(0.0);
		}
	}
	
	return &m_desired;
}// MyActionLimit::fire
