// ProbabilityMap.h

#ifndef My__ProbabilityMap
#define My__ProbabilityMap


// Dependencies

#include "vec2.h"
#include "Ray.h"
#include <set>


// PathPoint struct
struct PathPoint {
	inline PathPoint() {}
	inline PathPoint(unsigned int xx, unsigned int yy, unsigned int c) : x(xx), y(yy), cost(c) {}
	
	unsigned int	x, y;
	unsigned int	cost;
};

// A rect..
struct Rect {
	int				left, top, width, height;
};

// Comparator for PathPoints
struct lpathpoint {
	bool operator()(const PathPoint &p0, const PathPoint &p1) const {
		return (p0.cost <= p1.cost);
	}
};

// ProbabilityMap class
class ProbabilityMap {
	public:
		// ProbabilityMap.cpp
		ProbabilityMap(int w, int h);
		~ProbabilityMap();
		
		// Sonar input
		void input(const vec2d &pos, const Ray *rays, int nRays);
		void addDirtyRect(const Rect &nrect);
		Rect getTotalRect() const;
		
		// Info
		int width() const;
		int height() const;
		vec2d mapCenter() const;
		
		
		// ProbabilityMapImage.cpp
		// Map image data (for viewing)
		bool getMapData(char *data, int *lena) const;
		bool getTempData(char *data, int *lena) const;
		
		
		// ProbabiltyMapPath.cpp
		// Pathfinding and occupancy
		bool cellOpen(unsigned int x, unsigned int y);
		void addPoint(std::set<PathPoint, lpathpoint> &list, unsigned int x, unsigned int y, unsigned int cost, unsigned int px, unsigned int py);
		void makePath(unsigned int x, unsigned int y, vec2d *path, int *pathLena);
		int getPath(int *path);
		void discardPath();
		void resetCosts();
		bool findPathToUnknown(const vec2d &pos, vec2d *path, int *pathLena);
		bool findPathToTarget(const vec2d &pos, const vec2d &target, vec2d *path, int *pathLena);
		bool isBlocked(const vec2d &p0, const vec2d &dir, int steps, int width);
		
		
		// ProbabilityMapFill.cpp
		// Cell filling routines
		void fillRect(const vec2i &center, const vec2i &size, unsigned int val);
		void fillRect(const vec2i &center, const vec2i &size, unsigned int val, unsigned int clipVal);
		void fillRect2(const vec2d &p0, const vec2d &p1, const vec2d &size, unsigned int val);
		void fillLine2(const Ray &ray0, const Ray &ray1, double width, unsigned int weight, unsigned int minVal);
		void fillPoly(const vec2d *points, int nPoints, unsigned int weight, unsigned int minVal);
		void fillPoly(const vec2d *points, int nPoints, unsigned int val);
	
	
	private:
		// Cells
		unsigned int	*m_cells;
		unsigned int	*m_tempCells;
		unsigned int	*m_costs;
		unsigned int	*m_prev;
		int				m_width;
		int				m_height;
		
		Rect			m_totalRect;
		bool			m_hasTotalRect;
		
		// Input sonar rays
		Ray				m_rays[8];
		bool			m_hasRays;
		
		// Calibration settings
		vec2i			m_nearSize;
		vec2i			m_veryNearSize;
		double			m_rayWidth;
		double			m_rayRetract;
		vec2i			m_hitSize;
		vec2i			m_largeHitSize;
		
		// Pathfinding
		unsigned int	m_maxPassableCost;
		
		vec2i			*m_path;
		int				m_pathLen;
};


#endif
