// ProbabilityMapImage.cpp

// Member functions of ProbabilityMap to get an image of the map,
// for viewing purposes

#include "ProbabilityMap.h"



// Write map data to bitmap image
// Assumes 'data' is a large enough memory block to hold the image
bool ProbabilityMap::getMapData(char *data, int *lena) const {
	if(!m_hasTotalRect) return false;
	
	int left = m_totalRect.left;
	int top = m_totalRect.top;
	int width = m_totalRect.width;
	int height = m_totalRect.height;
	
	// Need multiple of 4 width for bitmaps
	// (map dimensions must be a multiple of 4 for this to be truly safe)
	if((width%4) != 0)
		width += (4 - (width%4));
	
	// Bmp header
	data[0] = 0x42;
	data[1] = 0x4d;
	
	*(unsigned int*)(data+2) = 26+height*width*3;
	*(unsigned int*)(data+10) = 26;
	*(unsigned int*)(data+14) = 12;
	*(unsigned short*)(data+18) = width;
	*(unsigned short*)(data+20) = height;
	*(unsigned short*)(data+22) = 1;
	*(unsigned short*)(data+24) = 24;
	
	// Image data
	char *idata = data+26;
	for(int i=top;i<top+height;i++) {
		for(int j=left;j<left+width;j++) {
			int ic = *(m_cells + i*m_width + j) & 0xff;
			if(ic > 255) ic = 255;
			char c = (char)ic;
			//char c = *(char*)(m_cells + i*m_width + j);
			*idata++ = c;
			*idata++ = c;
			*idata++ = c;
		}
	}
	
	// Draw last detected path to the image
	for(int i=0;i<m_pathLen;i++) {
		const vec2i &p(m_path[i]);
		
		if(p.x >= left && p.y >= top && p.x < left+width && p.y < top+height) {
			data[26 + (p.y-top)*width*3 + (p.x-left)*3 + 0] = 0;
			data[26 + (p.y-top)*width*3 + (p.x-left)*3 + 1] = -1;
			data[26 + (p.y-top)*width*3 + (p.x-left)*3 + 2] = -1;
		}
	}
	
	*lena = 26+height*width*3;
	
	return true;
}// ProbabilityMap::getMapData


// Write temp map data to bitmap image
// (This is just the points recorded from the sonars, for debug viewing)
bool ProbabilityMap::getTempData(char *data, int *lena) const {
	if(!m_hasTotalRect) return false;
	
	int left = m_totalRect.left;
	int top = m_totalRect.top;
	int width = m_totalRect.width;
	int height = m_totalRect.height;
	
	// Need multiple of 4 width for bitmaps
	// (map dimensions must be a multiple of 4 for this to be truly safe)
	if((width%4) != 0)
		width += (4 - (width%4));
	
	// Bmp header
	data[0] = 0x42;
	data[1] = 0x4d;
	
	*(unsigned int*)(data+2) = 26+height*width*3;
	*(unsigned int*)(data+10) = 26;
	*(unsigned int*)(data+14) = 12;
	*(unsigned short*)(data+18) = width;
	*(unsigned short*)(data+20) = height;
	*(unsigned short*)(data+22) = 1;
	*(unsigned short*)(data+24) = 24;
	
	// Image data
	char *idata = data+26;
	for(int i=top;i<top+height;i++) {
		for(int j=left;j<left+width;j++) {
			char c = *(char*)(m_tempCells + i*m_width + j);
			*idata++ = c;
			*idata++ = c;
			*idata++ = c;
		}
	}
	
	*lena = 26+height*width*3;
	
	return true;
}// ProbabilityMap::getTempData
